import queue
import threading
import time
from random import uniform,shuffle

from .shots import PORT_UNKNOWN,PORT_FILTERED,PORT_OPEN
from .shots.even import EVENShot
from .shots.even6 import EVEN6Shot
from .shots.rprn import RPRNShot
from .shots.rrp import RRPShot

from . import logger

MAX_QUEUE_SIZE = 5000
TARGET_PORTS = [135, 445, 3389, 5985, 5986]

DEFAULT_SHOTS = [EVENShot, RPRNShot, RRPShot, EVEN6Shot]

class CornerShot(object):
    def __init__(self, username, password, domain, workers=250, shots=None):

        logger.debug(f'CS created with username: {username},domain:{domain},workers:{workers}')
        if shots is None:
            shots = DEFAULT_SHOTS

        self.shot_classes = shots
        self.username = username
        self.password = password
        self.domain = domain
        self.workers = workers
        self.bulletQ = queue.Queue()
        self.resultQ = queue.Queue()
        self.runthreads = True
        self.results = {}
        self.shot_list = []
        self.total_shots = 0
        self.current_tasks = []
        self.skip_scanned = False
        self.already_scanned = []
        self.batch_scanned_event = None

    def _takeashot(self):
        while self.runthreads:
            res = None
            try:
                bullet = self.bulletQ.get(timeout=0.1)
                if bullet:
                    try:
                        res = bullet.shoot()
                    except TimeoutError:
                        logger.debug(f'Timeout error')
                    except Exception:
                        logger.debug(f'Unexpected exception during shot', exc_info=True)
                    finally:
                        self.bulletQ.task_done()
                        self.resultQ.put(res)
                logger.debug(f'Bullet is none!')
            except (queue.Empty,TimeoutError):
                logger.debug(f'Exception during bullet load', exc_info=True)
            except Exception:
                logger.debug(f'Unexpected exception during bullet load', exc_info=True)

    def add_shots(self, destinations, targets, target_ports=None, destination_ports=None):

        if target_ports is None:
            target_ports = TARGET_PORTS

        self._shots_generator(destinations, targets, target_ports, destination_ports)

    def add_many_shot_pairs(self, carrier_target_pairs, target_ports=None, destination_ports=None):
        if target_ports is None:
            target_ports = TARGET_PORTS

        tport_shot_class = []
        for target_port in target_ports:
            tport_shot_class.append([target_port,self._get_suitable_shots(target_port, destination_ports)])

        for ct_pair in carrier_target_pairs:
            carrier = ct_pair[0]
            target = ct_pair[1]
            for tport_shot_class_pair in tport_shot_class:
                target_port = tport_shot_class_pair[0]
                for cls in tport_shot_class_pair[1]:
                    self.shot_list.append(cls(self.username, self.password, self.domain, carrier, target,target_port=target_port))

        self.total_shots += len(self.shot_list)

    def _shots_generator(self, destinations, targets, target_ports, destination_ports=None):
        for destination in destinations:
            for target in targets:
                for target_port in target_ports:
                    for cls in self._get_suitable_shots(target_port, destination_ports):
                        self.shot_list.append(cls(self.username, self.password, self.domain, destination, target,target_port=target_port))

    def _merge_result(self, dest, target, tport, state):
        if self.skip_scanned and PORT_OPEN in state:
            if target not in self.already_scanned:
                self.already_scanned.append(target)

        if dest not in self.results:
            self.results[dest] = {}
        if target not in self.results[dest]:
            self.results[dest][target] = {}
        if tport not in self.results[dest][target]:
            self.results[dest][target][tport] = state
        elif PORT_UNKNOWN in self.results[dest][target][tport]:
            self.results[dest][target][tport] = state
        elif PORT_FILTERED in self.results[dest][target][tport]:
            pass
        elif PORT_FILTERED in state:
            self.results[dest][target][tport] = state
        elif PORT_UNKNOWN in state:
            pass
        elif state not in self.results[dest][target][tport]:
            self.results[dest][target][tport] += "|" + state

    def _get_next_tasks(self,remaining):

        new_tasks = []
        remaining = min(len(self.shot_list),remaining)
        if self.skip_scanned:
            iterated_shots = 0
            for bullet in self.shot_list:
                if remaining > 0:
                    if bullet.target not in self.already_scanned:
                        new_tasks.append(bullet)
                    else:
                        logger.info(f"Skipping {bullet.target}, already scanned...")
                        self.total_shots -= 1
                    iterated_shots += 1
                    remaining -= 1
                else:
                    break
            self.shot_list = self.shot_list[iterated_shots:]

        else:
            new_tasks = self.shot_list[0:remaining ]
            self.shot_list = self.shot_list[remaining + 1:]

        return new_tasks

    def _shots_manager(self):
        remaining = min(self.workers,MAX_QUEUE_SIZE)
        shuffle(self.shot_list)
        self.total_shots = len(self.shot_list)

        while self.runthreads:
            self.current_tasks = self._get_next_tasks(remaining)

            remaining = remaining - len(self.current_tasks)

            for bt in self.current_tasks:
                self.bulletQ.put(bt)

            while True:
                try:
                    result = self.resultQ.get(timeout=5)
                    if result:
                        destination, target, target_port, state = result
                        self._merge_result(destination, target, target_port, state)
                    self.resultQ.task_done()
                    remaining += 1
                    self.total_shots -= 1
                    if self.total_shots % 500 == 0:
                        self.batch_scanned_event.set()
                    if self.total_shots < 1:
                        self.runthreads = False
                except (TimeoutError,queue.Empty):
                    break

        self.total_shots = 0
        self.batch_scanned_event.set()

    def open_fire(self,blocking=True,skip_scanned=False):
        self.skip_scanned = skip_scanned
        self.total_shots = len(self.shot_list)

        self.workers = min(self.total_shots,self.workers)

        if self.total_shots > 0:
            for _ in range(self.workers):
                w = threading.Thread(target=self._takeashot, daemon=True)
                w.start()

        if blocking:
            self._shots_manager()
            return self.results
        else:
            main_thread = threading.Thread(target=self._shots_manager,daemon=True)
            main_thread.start()
            self.batch_scanned_event = threading.Event()
            return self.batch_scanned_event

    def read_results(self):
        return self.results

    def remaining_shots(self):
        return self.total_shots

    def _get_suitable_shots(self, target_port, destination_port):
        class_list = []
        for bc in self.shot_classes:
            if destination_port:
                if destination_port in bc.destination_port_range() and target_port in bc.target_port_range():
                    class_list.append(bc)
            elif target_port in bc.target_port_range():
                class_list.append(bc)

        return class_list




from impacket.dcerpc.v5 import rrp

from . import *

TS = ('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0')
IFACE_UUID = rrp.MSRPC_UUID_RRP


class RRPShot(BaseRPCShot):

    def __init__(self, username, password, domain, destination, target, dest_port=None, target_port=None):
        BaseRPCShot.__init__(self, username, password, domain, TS, IFACE_UUID, destination, target, dest_port,
                             target_port,
                             state_exception_open=['rpc_s_access_denied', 'ERROR_BAD_NET_NAME'],
                             state_exception_closed=['ERROR_BAD_NETPATH'])

    @staticmethod
    def target_port_range():
        return [445]

    @staticmethod
    def destination_port_range():
        return [445]

    def do_binding(self):
        return r'ncacn_np:%s[\PIPE\winreg]' % self.destination

    def do_rpc_logic(self):
        try:
            resp = rrp.hOpenCurrentUser(self.dce)
        except Exception:
            return

        rrp.hBaseRegSaveKey(self.dce, resp['phKey'], f'\\\\{self.target}\\shmores\\file')



from argparse import ArgumentParser
import re
from ipaddress import ip_network, AddressValueError, NetmaskValueError, summarize_address_range, ip_address,collapse_addresses
from .cornershot import CornerShot
from . import logger
import logging
from json import dumps

DEFAULT_NUM_THREADS = 200
DEFAULT_TARGET_PORTS = [135, 445, 3389, 5985, 5986]

INVALID_SUBNET_ERROR_MESSAGE = f"please pick a valid comma delimited list of ip subnet or range such as '192.168.10.0/24,10.9.0.0-10.9.0.255'"
INVALID_PORTS_ERROR_MESSAGE = f"please pick a valid comma delimited list of port ranges, or list of ports"


def parse_args():
    parser = ArgumentParser(prog="CornerShot", prefix_chars="-/", add_help=False, description=f'Corner Shooter')

    parser.add_argument('-h', '--help', '/?', '/h', '/help', action='help', help='show this help message and exit')
    parser.add_argument("user", help="provide any authenticated user in the domain", type=str)
    parser.add_argument("password", help="domain password", type=str)
    parser.add_argument("domain", help="the FQDN of the domain.", type=str)
    parser.add_argument("carrier", help="carrier host for cornershot", type=str)
    parser.add_argument("target", help="target for shot", type=str)
    parser.add_argument("-tp", "--tports", dest='tports', default=DEFAULT_TARGET_PORTS, help="comma delimited list of target port ranges to scan for", type=str)
    parser.add_argument("-w", "--workerthreads", dest='threads', help="number of threads to perform shots", default=DEFAULT_NUM_THREADS, type=int)
    parser.add_argument('-v', dest='verbose', action='store_true', help='enable verbose logging')

    args = parser.parse_args()

    return args


def set_logger(is_verbose):
    log_level = logging.DEBUG if is_verbose else logging.INFO

    logger.setLevel(log_level)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)

    console_formatter = logging.Formatter('%(message)s')
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)


def parse_ip_ranges(ip_ranges):
    add_list = []
    try:
        if ip_ranges:
            if ',' in ip_ranges:
                ip_ranges = ip_ranges.split(',')
            else:
                ip_ranges = [ip_ranges]

            for address in ip_ranges:
                if '-' in address:
                    first_addr = ip_address(address.split('-')[0])
                    last_addr = ip_address(address.split('-')[1])
                    for ipnet in summarize_address_range(ip_address(first_addr), ip_address(last_addr)):
                        add_list += [ipad.exploded for ipad in ipnet]
                elif '/' in address:
                    add_list += [ipad.exploded for ipad in ip_network(address, strict=False)]
                else:
                    add_list.append(address)

    except (AddressValueError, NetmaskValueError, ValueError):
        raise ValueError(INVALID_SUBNET_ERROR_MESSAGE)

    return add_list


def parse_port_ranges(ranges):
    port_ranges = []

    if type(ranges) is list:
        if all(isinstance(x, int) and (0 < x < 65536) for x in ranges):
            return ranges
        else:
            raise ValueError(INVALID_PORTS_ERROR_MESSAGE)

    if ranges:
        if ',' in ranges:
            ranges = ranges.split(',')
        else:
            ranges = [ranges]

        for port_range in ranges:
            res = re.findall(r"(\d+)-?(\d*)", port_range)
            if res:
                pstart = int(res[0][0])
                pend = None if not res[0][1] else int(res[0][1])
                if pend:
                    port_ranges += [p for p in range(pstart, pend)]
                else:
                    port_ranges.append(pstart)
            else:
                raise ValueError(INVALID_PORTS_ERROR_MESSAGE)
    else:
        raise ValueError(INVALID_PORTS_ERROR_MESSAGE)

    return port_ranges


if __name__ == '__main__':
    try:
        cs = None
        args = parse_args()
        set_logger(args.verbose)
        logger.info('CornerShot starting...')

        cs = CornerShot(args.user, args.password, args.domain, workers=args.threads)
        cs.add_shots(parse_ip_ranges(args.carrier), parse_ip_ranges(args.target), target_ports=parse_port_ranges(args.tports))
        cs.open_fire()

    except KeyboardInterrupt:
        logger.info("Interrupted!")
    except Exception as err:
        logger.error(f"CornerShot got exception - {err}")
        logger.debug(f"CornerShot unexpected exception!", exc_info=True)
    finally:
        logger.info('CornerShot finished...')
        if cs:
            res = cs.read_results()
            if res: logger.info(dumps(res,indent=4))



from impacket.dcerpc.v5 import rprn
from impacket.dcerpc.v5.dtypes import NULL

from . import *

TS = ('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0')
IFACE_UUID = rprn.MSRPC_UUID_RPRN


class RPRNShot(BaseRPCShot):

    def __init__(self, username, password, domain, destination, target, dest_port=None, target_port=None):
        BaseRPCShot.__init__(self, username, password, domain, TS, IFACE_UUID, destination, target, dest_port,
                             target_port,state_exception_closed=['INVALID_PRINTER_NAME'])

    @staticmethod
    def target_port_range():
        yield from range(1, 445)
        yield from range(446, 65535 + 1)

    @staticmethod
    def destination_port_range():
        return [445]

    def do_binding(self):
        return r'ncacn_np:%s[\PIPE\spoolss]' % self.destination

    def _create_request(self):
        request = rprn.RpcOpenPrinter()
        str = f"http://{self.target}:{self.trgt_port}/printers/ppp/.printer"

        request['pPrinterName'] = '%s\x00' % str
        request['pDatatype'] = NULL
        request['pDevModeContainer']['pDevMode'] = NULL
        request['AccessRequired'] = rprn.SERVER_READ

        return request

    def do_rpc_logic(self):
        request = self._create_request()
        self.dce.request(request)



from impacket.dcerpc.v5 import even
from impacket.dcerpc.v5.dtypes import NULL

from . import *

TS = ('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0')
IFACE_UUID = even.MSRPC_UUID_EVEN

class EVENShot(BaseRPCShot):

    def __init__(self, username, password, domain, destination, target, dest_port=None, target_port=None):
        BaseRPCShot.__init__(self, username, password, domain, TS, IFACE_UUID, destination, target, dest_port,
                             target_port,
                             state_exception_open=['BAD_NETWORK_NAME', 'NAME_NOT_FOUND', 'ACCESS_DENIED'],
                             state_exception_closed=['BAD_NETWORK_PATH'])

    @staticmethod
    def target_port_range():
        return [445]

    @staticmethod
    def destination_port_range():
        return [445]


    def do_binding(self):
        return r'ncacn_np:%s[\PIPE\eventlog]' % self.destination

    def _create_request(self):
        request = even.ElfrOpenBELW()

        str = f"\\??\\UNC\\{self.target}\\share\\file"

        if self.trgt_port != 445:
            # Will only work if WebDav service (WebClient) is running on client OS
            str = f'\\??\\UNC\\{self.target}@{self.trgt_port}\\DavWWWRoot\\share\\file'

        request['UNCServerName'] = NULL
        request['BackupFileName'] = str
        request['MajorVersion'] = 1
        request['MinorVersion'] = 1

        return request

    def do_rpc_logic(self):
        request = self._create_request()
        self.dce.request(request)


from impacket.dcerpc.v5 import even6, epm
from impacket.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_PRIVACY

from . import *

TS = ('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0')
IFACE_UUID = even6.MSRPC_UUID_EVEN6


class EVEN6Shot(BaseRPCShot):

    def __init__(self, username, password, domain, destination, target, dest_port=None, target_port=None):
        BaseRPCShot.__init__(self, username, password, domain, TS, IFACE_UUID, destination, target, dest_port,
                             target_port, auth_level=RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
                             state_exception_closed=["BAD_NETPATH"], state_exception_open=['ERROR_FILE_NOT_FOUND'])

    @staticmethod
    def target_port_range():
        return [445]

    @staticmethod
    def destination_port_range():
        return [135]

    def do_binding(self):
        return epm.hept_map(self.destination, even6.MSRPC_UUID_EVEN6, protocol='ncacn_ip_tcp')

    def _create_request(self):
        request = even6.EvtRpcOpenLogHandle()

        request['Channel'] = f'\\\\{self.target}\\share\\file\x00'
        request['Flags'] = 2  # specifies a file name

        return request

    def do_rpc_logic(self):
        request = self._create_request()
        self.dce.request(request)
